Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type Interfaces.CHAR_DECODER
Decoder for unicode characters.
A character decoder has some valid initial value init. In order to decode a character bytes has to be pushed into the decoder by put until the decoder is either complete (is_complete) or has encoutered some decoding error (has_error).
If the decoding is complete the decoded unicode character is returned by uchar. As long as the scan is not complete or if the scan has encountered some decoding error the function uchar returns rep = U+FFFD.
type t
Partially or completely scanned unicode character.