Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type Interfaces.MINIMAL_PARSER
A minimal parser is a sink of tokens which either succeeds or returns a list of failed syntax expectations.
A parser p is a sink of token. As long as it signals needs_more p more token can be pushed into the parser via put token p or the input stream can be ended via put_end p.
has_ended p is equivalent to not (needs_more p). has_ended p signals that the parser has either succeeded or failed.
If it has succeeded the final value is available via final p.